unit uSorts;

interface

uses
  SysUtils;

type
  { Types numr des diffrents algorithmes possibles. }
  TAlgorithm = (aQuickSort, aShearSort);

  { Types-mthodes callback. }
  TCompareFunction = function (const Item1, Item2: Integer): Integer;
  TExchangeProc = procedure (const Item1, Item2: Integer);

  TObjCompareFunction = function (const Item1, Item2: Integer): Integer of object;
  TObjExchangeProc = procedure (const Item1, Item2: Integer) of object;
  { Classe abstraite grant les comparaisons et les changes. }
  TSort = class(TObject)
  private
    function CallCompare(const Item1, Item2: Integer): Integer; virtual;
            abstract;
    procedure CallExchange(const Item1, Item2: Integer); virtual; abstract;
  end;

  { Tri avec des mthodes appartenant  des objets. }
  TObjSort = class(TSort)
  private
    FCompare: TObjCompareFunction;
    FExchange: TObjExchangeProc;
    function CallCompare(const Item1, Item2: Integer): Integer; override;
    procedure CallExchange(const Item1, Item2: Integer); override;
  public
    constructor Create(ACompareFunc: TObjCompareFunction; AExchangeProc:
            TObjExchangeProc);
  end;

  { Tri avec des mthodes n'appartenant pas  des objets. }
  TClassicSort = class(TSort)
  private
    FCompare: TCompareFunction;
    FExchange: TExchangeProc;
    function CallCompare(const Item1, Item2: Integer): Integer; override;
    procedure CallExchange(const Item1, Item2: Integer); override;
  public
    constructor Create(ACompareFunc: TCompareFunction; AExchangeProc:
            TExchangeProc);
  end;

  { Classe abstaite grant un tri. }
  TSortAlgorithm = class(TObject)
  private
    FSort: TSort;
  public
    constructor Create(ACompareFunc: TCompareFunction; AExchangeProc:
            TExchangeProc); overload;
    constructor Create(ACompareFunc: TObjCompareFunction; AExchangeProc:
            TObjExchangeProc); overload;
    destructor Destroy; override;
    procedure Sort(const AMin, AMax: Integer); virtual; abstract;
  end;

  TQuickSort = class(TSortAlgorithm)
  public
    procedure Sort(const AMin, AMax: Integer); override;
  end;

  TShearSort = class(TSortAlgorithm)
  private
    FCols: Integer;
//    FLog: Integer;
    FMax: Integer;
    FMin: Integer;
    FRows: Integer;
    function AlteredCompare(Item1, Item2: Integer): Integer;
    procedure AlteredExchange(Item1, Item2: Integer);
    procedure RealSort(Length: Integer);
    procedure SortPart1(Lo, Hi, Nx: Integer; Up: Boolean);
    procedure SortPart2(Lo, Hi, Nx: Integer; Up: Boolean);
  public
    procedure Sort(const AMin, AMax: Integer); override;
  end;


procedure SortAnything(const AMin, AMax: Integer; ACompareFunc:
        TCompareFunction; AExchangeProc: TExchangeProc; Algorithm: TAlgorithm);
        overload;

procedure SortAnything(const AMin, AMax: Integer; ACompareFunc:
        TObjCompareFunction; AExchangeProc: TObjExchangeProc; Algorithm:
        TAlgorithm); overload;

implementation

procedure SortAnything(const AMin, AMax: Integer; ACompareFunc:
        TCompareFunction; AExchangeProc: TExchangeProc; Algorithm: TAlgorithm);
var
  Sort: TSortAlgorithm;
begin
  Assert(AMax > AMin, 'AMax doit tr suprieur  AMin.');
  case Algorithm of
    aQuickSort: Sort := TQuickSort.Create(ACompareFunc, AExchangeProc);  
    aShearSort: Sort := TShearsort.Create(ACompareFunc, AExchangeProc);  
  else
    raise Exception.Create('Algorithme non pris en charge.');
  end; // case.
  try
    Sort.Sort(AMin, AMax);
  finally
    Sort.Free;
  end; // try.
end;

procedure SortAnything(const AMin, AMax: Integer; ACompareFunc:
        TObjCompareFunction; AExchangeProc: TObjExchangeProc; Algorithm:
        TAlgorithm);
var
  Sort: TSortAlgorithm;
begin
  Assert(AMax > AMin, 'AMax doit tr suprieur  AMin.');
  case Algorithm of
    aQuickSort: Sort := TQuickSort.Create(ACompareFunc, AExchangeProc);  
    aShearSort: Sort := TShearsort.Create(ACompareFunc, AExchangeProc);  
  else
    raise Exception.Create('Algorithme non pris en charge.');
  end; // case.
  try
    Sort.Sort(AMin, AMax);
  finally
    Sort.Free;
  end; // try.  
end;

{
*********************************** TObjSort ***********************************
}
constructor TObjSort.Create(ACompareFunc: TObjCompareFunction; AExchangeProc:
        TObjExchangeProc);
begin
  inherited Create;
  FCompare  := ACompareFunc;
  FExchange := AExchangeProc;
end;

function TObjSort.CallCompare(const Item1, Item2: Integer): Integer;
begin
  Result := FCompare(Item1, Item2);
end;

procedure TObjSort.CallExchange(const Item1, Item2: Integer);
begin
  FExchange(Item1, Item2);
end;

{
********************************* TClassicSort *********************************
}
constructor TClassicSort.Create(ACompareFunc: TCompareFunction; AExchangeProc:
        TExchangeProc);
begin
  inherited Create;
  FCompare  := ACompareFunc;
  FExchange := AExchangeProc;
end;

function TClassicSort.CallCompare(const Item1, Item2: Integer): Integer;
begin
  Result := FCompare(Item1, Item2);
end;

procedure TClassicSort.CallExchange(const Item1, Item2: Integer);
begin
  FExchange(Item1, Item2);
end;

{
******************************** TSortAlgorithm ********************************
}
constructor TSortAlgorithm.Create(ACompareFunc: TCompareFunction;
        AExchangeProc: TExchangeProc);
begin
  inherited Create;
  FSort := TClassicSort.Create(ACompareFunc, AExchangeProc);
end;

constructor TSortAlgorithm.Create(ACompareFunc: TObjCompareFunction;
        AExchangeProc: TObjExchangeProc);
begin
  inherited Create;
  FSort := TObjSort.Create(ACompareFunc, AExchangeProc);
end;

destructor TSortAlgorithm.Destroy;
begin
  FSort.Free;
  inherited;
end;

{
********************************** TQuickSort **********************************
}
procedure TQuickSort.Sort(const AMin, AMax: Integer);
var
  Lo, Hi, Mid: Integer;
begin
  { Affectation. }
  Lo  := AMin;
  Hi  := AMax;
  Mid := (Lo + Hi) div 2;

  { Boucle. }
  repeat

    { Comparaisons. }
    while FSort.CallCompare(Lo, Mid) < 0 do
      Inc(Lo);
    while FSort.CallCompare(Hi, Mid) > 0 do
      Dec(Hi);

    if Lo <= Hi then
    begin
      { Echange seulement si les index sont diffrents. }
      if Lo <> Hi then
      begin
        { Change l'index du "milieu" s'il est gal  Lo ou Hi. }
        if Mid = Lo then
          Mid := Hi
        else if Mid = Hi then
          Mid := Lo;

        { Echange les deux cases du tableau. }
        FSort.CallExchange(Lo, Hi);
      end;
      Inc(Lo);
      Dec(Hi);
    end;

  until Lo > Hi;

  { Appels rcursifs. }
  if Hi > AMin then
    Sort(AMin, Hi);
  if Lo < AMax then
    Sort(Lo, AMax);
end;

{
********************************** TShearSort **********************************
}
function TShearSort.AlteredCompare(Item1, Item2: Integer): Integer;
begin
  Result := FSort.CallCompare(Item1 + FMin, Item2 + FMin);
end;

procedure TShearSort.AlteredExchange(Item1, Item2: Integer);
begin
  FSort.CallExchange(Item1 + FMin, Item2 + FMin);
end;

procedure TShearSort.RealSort(Length: Integer);
var
  Pow, Log: Integer;
  Divi: Integer;
  I, J, K: Integer;
begin
  Pow  := 1;
  Divi := 1;

  I := 1;
  while I * I <= Length do
  begin
    if I mod Length = 0 then
      Divi := I;
    Inc(I);
  end;

  Log := 0;
  while Pow <= FRows do
  begin
    Pow := Pow * 2;
    Inc(Log);
  end;

  FRows := Divi;
  FCols := Length div Divi;

  for K := 0 to Log -1 do
  begin
    J := 0;
    while J < FCols div 2 do
    begin
      for I := 0 to FRows -1 do
        SortPart1(I * FCols, (I+1) * FCols, 1, I mod 2 = 0);
      for I := 0 to FRows -1 do
        SortPart2(I * FCols, (I+1) * FCols, 1, I mod 2 = 0);
      Inc(J);
    end;

    J := 0;
    while J < FRows div 2 do
    begin
      for I := 0 to FCols -1 do
        SortPart1(I, FRows * FCols + I, FCols, True);
      for I := 0 to FCols -1 do
        SortPart2(I, FRows * FCols + I, FCols, True);
      Inc(J);
    end;
  end;

  J := 0;
  while J < FCols div 2 do
  begin
    for I := 0 to FRows -1 do
      SortPart1(I * FCols, (I+1) * FCols, 1, True);
    for I := 0 to FRows -1 do
      SortPart2(I * FCols, (I+1) * FCols, 1, True);
    Inc(J);
  end;
end;

procedure TShearSort.Sort(const AMin, AMax: Integer);
begin
  FMin  := AMin;
  FMax  := AMax;
  FCols := 0;
  Frows := 0;
  RealSort(FMax - FMin + 1);
end;

procedure TShearSort.SortPart1(Lo, Hi, Nx: Integer; Up: Boolean);
var
  J: Integer;
begin
  J := Lo;
  while J + Nx < Hi do
  begin
    if (Up and (AlteredCompare(J, J + Nx) > 0)) or
    (not Up and (AlteredCompare(J, J + Nx) < 0)) then
      AlteredExchange(J, J + Nx);

    J := J + 2 * Nx;
//    Inc(J, J + 2 * Nx);
  end;
end;

procedure TShearSort.SortPart2(Lo, Hi, Nx: Integer; Up: Boolean);
var
  J: Integer;
begin
  J := Lo + Nx;
  while J + Nx < Hi do
  begin
    if (Up and (AlteredCompare(J, J + Nx) > 0)) or
    (not Up and (AlteredCompare(J, J + Nx) < 0)) then
      AlteredExchange(J, J + Nx);

    J := J + 2 * Nx;
//    Inc(J, J + 2 * Nx);
  end;
end;


end.
